/*
*	Quork C preprocessor, version 0.1
*	Anicka Bernathova, <anicka@anicka.net>, 05/2007
*	License: GNU GPLv2
*	
*	lexical analyzer part	
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "quork.h"

#define MAXTOKEN 4095
#define OP2(x,y) ((x << 8) | y)
#define OUR_EOF -108
#define FILE_DEPTH_MAX 511

struct file {
	struct file* next;
	FILE* fp;
	int lastln;
	int l3_temp; /* we must save this one */
	char path[1];
};

/* global variables */
static int ln; /* current line number */

static int l2_temp; /* temporary variable for kill_comments() */
static int l3_temp; /* temporary variable for mk_token() */

static char* tok_buf; /* pointer to buffer storing text of processed token */
static int tok_size; /* current size of buf */
static int tok_count; /* counter for tok_buf */
static int tok_comments_on; /* switch on/off killing of comments (necessary for strings and chars) */
static struct file* file_stack; /* global file stack */
static int file_depth; /* current depth of nested includes */

int tok_inc_enable; /* switch from/to include tokenizer mode */
char* global_path;

/* generic error message */
void __attribute__((noreturn))
err(char* msg)
{
	if (file_stack)
		fprintf(stderr,"ERROR: %s (%d): %s\n",file_stack->path,ln,msg);
	else
		fprintf(stderr,"ERROR: %s\n",msg);
	exit(1);
}

/* generic warning message */
void
warn(char* msg)
{
	if (file_stack)
		fprintf(stderr,"WARNING: %s (%d): %s\n",file_stack->path,ln,msg);
	else
		fprintf(stderr,"WARNING: %s\n",msg);
}

void
print_file_stack(void)
{
	struct file* temp;

	temp=file_stack;
	while(temp) {
		printf("%s \n",temp->path);
		temp=temp->next;
	}
}

static struct file*
new_file(int length)
{
	struct file* pf;
	if (!(pf=malloc(sizeof(struct file)+length)))
		err("Low memory");
	return pf;

}

static void
push_file(struct file* fp)
{
	file_depth++;
	if (file_depth>FILE_DEPTH_MAX)
		err("Too many nested includes");	
	fp->next=file_stack;
	file_stack=fp;
}

static void
pop_file(void)
{
	struct file* temp;
	
	file_depth--;
	temp=file_stack;
	if (file_stack){
		fclose(file_stack->fp);
		file_stack=file_stack->next;
		free(temp);
	}
	if (file_stack){
		ln=file_stack->lastln;
		l3_temp=file_stack->l3_temp;
	}
		
}

/*
common tasks for opening included file
*/
static void
prep_switch(char* path)
{
	struct file* nf;
	FILE* fp;

	fp=fopen(path,"r");
        if (fp==NULL)
		err("File not found");
	nf=new_file(strlen(path)+1);
	strcpy(nf->path,path);
	nf->fp=fp;
	nf->next=NULL;
	if (file_stack){
		file_stack->lastln=ln;
		file_stack->l3_temp=l3_temp;
	}
	ln=1;
	l3_temp=0;
	push_file(nf);
}

/*
open new include and push it to the top of the file stack
*/
void
switch_file(char* text, enum ttype t)
{
	char* buf;
	char* path;

	switch (t) {
		case FILENAME:
			/*
			C99 says this is implementation defined
			while we do not have ifdef, using standard headers
			is nearly impossible anyway - so we can do anything...
			*/
			path="/usr/include";
					break;
		case STRING:
			path=global_path;
			break;
		default:
			err("This error should never appear :-D");	
			break;
		}	
			if (!(buf=malloc(strlen(path)+strlen(text)+3)))
				err("Low memory");
			strcpy(buf,path);
			strcat(buf,"/");
			strcat(buf,text);
			prep_switch(buf);

}

/* reads char from file, counts lines: feed for kill_comments()  */
static int
rcc(void)
{
	int ret;
	if (!file_stack) /* we must be able to return EOF more than once */
		return EOF;
	ret=fgetc(file_stack->fp);
	if (ret=='\0'){
		warn("Weird \\0 in input.");
		return rcc();
	}
	if (ret=='\n')
		ln++;

	if(ret<0){
		pop_file();
		if (file_stack)
			ret=OUR_EOF;
		else
			ret=EOF;
	}
	return ret;
}

/* layer one ungetc (two more to continue) */
static void
back(int c)
{	
	if (c<0)
		return;	
	ungetc(c,file_stack->fp);
	if (c=='\n')
		ln--;
	
}

/* join logical lines, prepare feed for tokenizer */
static int
kill_backslashes(void)
{
	int c;

	if ((c=rcc())!='\\')
		return c;
	else {
		if ((c=rcc())=='\n')
			return kill_backslashes();
		else {	
			back(c);	
			return '\\';
		}
			
	}
	
}

/* take char from temporary variable (layer 2 ungetc) or ask for new */
static int
l2_new(void)
{
	int ret;
	
	if (l2_temp){
		ret=l2_temp;
		l2_temp=0;
		return ret;
	} else {
		return kill_backslashes();
	}

}

static void
l2_back(int c)
{
	l2_temp=c;
}



/* substitutes comments for spaces*/
static int
kill_comments(void)
{
	int cur;

	cur=l2_new();	
	if(!tok_comments_on)
		return (cur==OUR_EOF?'\n':cur);

	if (cur!='/')
		return (cur==OUR_EOF?'\n':cur);
	
	if((cur=l2_new())=='/'){
		while ((cur=l2_new())!='\n' && cur>=0);
		return '\n';
	}
		
	if(cur!='*'){
		l2_back(cur);
		return '/';
	}
	
	while ((cur=l2_new())>=0){
		if(cur=='*')
			if ((cur=l2_new())=='/')
				return ' ';
			else
				l2_back(cur);
	}
	err("Unterminated comment.");
}

/* fills new token with given information */
static struct token*
tok_generic(enum ttype t)
{
	struct token* pt;

	pt=tok_new(tok_count);
	pt->type=t;
	strcpy(pt->text,tok_buf);
	pt->gl=ln;
	pt->next=pt;
	pt->previous=pt;

	return pt;
}

/* bunch of functions for determining type of chars
 they can get EOF */
static int
is_alpha(int c)
{
	return (c>=0 && isalpha(c));
}

static int
is_alnum(int c)
{
	return (c>=0 && isalnum(c));
}

static int
is_space(int c)
{
	return (c>=0 && isspace(c));
}
static int
is_digit(int c)
{
	return (c>=0 && isdigit(c));
}

/* take char from temporary variable (layer 3 ungetc) or ask for new */
static int
l3_new(void)
{
	int ret;
	
	if (l3_temp){
		ret=l3_temp;
		l3_temp=0;
		return ret;
	} else {
		return kill_comments();
	}

}

static void
l3_back(int c)
{
	l3_temp=c;
}

/* allocate the tokenizer buffer or double its size */
static void
tok_prep_buf(void)
{
	if (!tok_buf){
		if(!(tok_buf=(malloc(MAXTOKEN))))
			err("Low memory");
		tok_size=MAXTOKEN;
	} else {
		if(!(tok_buf=(realloc(tok_buf,2*tok_size))))
			err("Low memory");
		tok_size*=2;
		
	}
		
}

/* adds character to token text */
static void
tok_add_char(int c)
{
	tok_buf[tok_count++]=c;
	if (tok_count>=tok_size)
		tok_prep_buf();
	tok_buf[tok_count]='\0';
}

/* reads input and tokenize it 
limitations: recognizing of digraphs not implemented
	    recognizing of operator ... not implemented
idea: with one notable exception (.) we can recognize
token type from the first character, so this function
is a giant switch.
*/
struct token*
mk_token(void)
{
	struct token* pt;
	int c;

	c=l3_new();

	if(c<0)
		return NULL;

	tok_count=0;

	if (is_alpha(c) || c=='_'){
		do {
			tok_add_char(c);
		} while(is_alnum(c=l3_new()) || c=='_');
		l3_back(c);
		pt=tok_generic(IDENTIFIER);
	} else if (is_space(c)) {
		int nl,pos;

		pos=-1;
		nl=(c=='\n');
		do {
			if (c=='\n'){
				if (!nl++)
					pos=tok_count;
				else
					tok_count=pos+1;
			}
			tok_add_char(c);
		} while(is_space(c=l3_new()));
		l3_back(c);
		pt=tok_generic(nl?WHITESPACENL:WHITESPACE);
	} else if (is_digit(c) || c=='.') {
		int temp;

		/* peep */
		temp=l3_new();
		l3_back(temp);	
		if(is_digit(temp) || is_digit(c)){
			int prev=0;

			do {
				prev=c;
				tok_add_char(c);
				c=l3_new();
			} while(is_alnum(c) ||
                               c=='.' ||
			       ((c=='+' || c=='-') &&
			         (prev=='e' || prev=='E' || prev=='p' || prev=='P')));
			l3_back(c);
			pt=tok_generic(NUMBER);
		} else {
			tok_add_char(c);
			pt=tok_generic(OPERATOR);
		}
	} else if (c=='#') {
		int d=0;

		tok_add_char(c);	
		if ((c=l3_new())=='#'){
			tok_add_char(c);
			d++;	
		} else {
			l3_back(c);
		}		
		pt=tok_generic(d?HASHHASH:HASH);
	} else if (c=='"') {
		tok_comments_on=0;
		do {
			tok_add_char(c);
			if ((c=l3_new())=='\\'){
				tok_add_char(c);
				if ((c=l3_new())=='\n')
					err("Unexpected \\n");
				tok_add_char(c);
				c=l3_new();
			}
		 	if (c=='\n')
				err("Unexpected \\n");
		} while(c!='"');
		tok_add_char(c);
		tok_comments_on=1;
		pt=tok_generic(STRING);
	} else if (c=='\'') {
		tok_comments_on=0;
		do {
			tok_add_char(c);
			if ((c=l3_new())=='\\'){
				tok_add_char(c);
				if ((c=l3_new())=='\n')
					err("Unexpected \\n");
				tok_add_char(c);
				c=l3_new();
			}
			if (c=='\n')
				err("Unexpected \\n");
		} while(c!='\'');
		tok_add_char(c);
		tok_comments_on=1;
		pt=tok_generic(CHAR);
	} else if (c=='(') {
		tok_add_char(c);	
		pt=tok_generic(LPAREN);
	} else if (c==')') {
		tok_add_char(c);	
		pt=tok_generic(RPAREN);
	} else if (c=='<' && tok_inc_enable) {
		do {
			tok_add_char(c);
			c=l3_new();
			if (c=='\n')
				err("Unexpected \\n");
		} while(c!='>');
		tok_add_char(c);
		pt=tok_generic(FILENAME);
	} else {
		int temp;

		temp=l3_new();
		switch (OP2(c,temp)){
		case OP2('!','='):
		case OP2('+','='):
		case OP2('-','='):
		case OP2('*','='):
		case OP2('/','='):
		case OP2('%','='):
		case OP2('>','='):
		case OP2('<','='):
		case OP2('|','='):
		case OP2('&','='):
		case OP2('^','='):
		case OP2('=','='):
		case OP2('-','>'):
		case OP2('+','+'):
		case OP2('-','-'):
		case OP2('&','&'):
		case OP2('|','|'):
			tok_add_char(c);
			tok_add_char(temp);
			break;
		case OP2('>','>'):
		case OP2('<','<'):
			tok_add_char(c);
			tok_add_char(temp);
			if ((c=l3_new())=='=')
				tok_add_char(c);
			else
				l3_back(c);	
			break;
		default:
			tok_add_char(c);
			l3_back(temp);	
			break;
		}
		pt=tok_generic(OPERATOR);
	}

	return pt;		
}

void
lex_init(char* filename)
{
	char* split, * fn;

	file_stack=NULL;
	tok_prep_buf();
	tok_inc_enable=0;
	tok_comments_on=1;
	/* FIXME: it would be kind to have a portability in mind, 
	this is unfortunately just a GNU extension */
	if (!(split=strrchr(filename,'/'))){
		global_path=".";
		fn=filename;
	} else {
		split[0]='\0';
		global_path=filename;
		fn=split+1;
	}
	switch_file(fn,STRING);
}
