/*
*	Quork C preprocessor, version 0.1
*	Anicka Bernathova, <anicka@anicka.net>, 05/2007
*	License: GNU GPLv2
*
* TODO:
*	bugfixes:
*		avoid substituting the same identifier more than once
*	planned features:
*		# operator, ## operator
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "quork.h"

#define HASHSIZE 103
#define MAGICC 17
#define NF -2

struct pp_macro {
	struct pp_macro* next;
	int arity;
	struct token* chain;
	char name[1];
};

struct pp_hash {
	struct pp_macro** hash;
	int nbuckets;
};

static struct pp_hash* defined_macros;

/* shows all known macros saved in a pp_hash structure */
static void
hash_show(struct pp_hash* hash)
{
	int i;
	struct pp_macro* pt;

	printf("	hash_show: Showing... ");
	for (i=0;i<hash->nbuckets;i++){
		pt=*(hash->hash+i);
		while (pt) {
			printf(" %s ",pt->name);
			pt=pt->next;
		}
	}
	printf(" ...finished\n");
}

/* determine hash bucket for the string, total number of buckets given*/
static int
hash_get_bucket(char* p, unsigned int nbuckets)
{
	unsigned int bucket;

	bucket=0;
	while (*p!='\0'){
		bucket=bucket * MAGICC + *p++;
	}
	bucket%=nbuckets;

	return bucket;

}

static struct pp_hash*
hash_new(void)
{
	struct pp_hash* new;
	int size;

	if (!(new=malloc(sizeof(struct pp_hash))))
		err("Low memory");
	size=HASHSIZE * sizeof(struct token*);
	if (!(new->hash=malloc(size)))
		err("Low memory!");
	memset(new->hash,0,size);
	new->nbuckets=HASHSIZE;

	return new;
}

/* return pointer to defined macro or null*/
static struct pp_macro*
hash_find(char* c, struct pp_hash* hash)
{
	struct pp_macro* pt;

	int i;
	i=hash_get_bucket(c,hash->nbuckets);

	pt=hash->hash[i];
	while (pt){
		if (!strcmp(pt->name,c))
			break;
		pt=pt->next;
	}
#ifdef DEBUG
	printf("	hash_find: We are resolving: %s and we found it:  %d\n",c,(pt?1:0));	
#endif
	return pt;
}

static void
hash_delete(char* name, struct pp_hash* hash)
{
	struct pp_macro* pt, ** ppt;

	ppt = &hash->hash[hash_get_bucket(name, hash->nbuckets)];
	while (pt = *ppt) {
		if (!strcmp(pt->name, name)) {
			*ppt = pt->next;
			free(pt);
			return;
		}
		ppt = &pt->next;
	}
}

static struct pp_macro* 
hash_insert(struct pp_macro* pm, struct pp_hash* hash)
{
	int bucket;
	struct pp_macro* pt, * retm, ** ppt;

	retm=NULL;
	bucket=hash_get_bucket(pm->name,hash->nbuckets);
	ppt=&hash->hash[bucket];
	while(pt=*ppt){
		if (!strcmp(pt->name,pm->name)){
			*ppt = pt->next;
			retm=pt;
			break;
		}
		ppt=&pt->next;
	}
	pm->next=pt;
	hash->hash[bucket]=pm;

	return retm;
}

static void
hash_free(struct pp_hash* hash)
{
	int i;
	struct pp_macro* pt, * temp;

	for (i=0;i<hash->nbuckets;i++){
		pt=*(hash->hash+i);
		while(pt) {
			temp=pt;
			pt=pt->next;
			free(temp);
		}
	}
	free(hash->hash);
	free(hash);
}

/* extracts a string between two delimiters 
this is used for parsing filenames of included headers
*/
static char*
extract_filename(char* c)
{
	char* ret;

	ret=c+1;
	while(*c!='\0'){
		c++;
	}
	c[-1]='\0';

	return ret;
}

static struct pp_macro*
new_macro(char *name, int no, struct token* chain)
{
	struct pp_macro* new;
	int l;

	l=strlen(name)+1;	
	if (!(new=malloc(sizeof(struct pp_macro)+l)))
		err("Low memory");
	
	strcpy(new->name,name);
	new->chain=chain;
	new->arity=no;
	new->next=NULL; /* this does not logically belong here but it safer this way */
	return new;
}

/*
saves arguments of a macro into the hash
known argument identifiers will be replaced for another token then
*/
static void
pp_save_arg(char* name, int no, struct pp_hash* hash)
{
	struct pp_macro* new;

	new=new_macro(name,no,NULL);
	hash_insert(new, hash);
}

/*
Prepares chain for saving: substitutes identifiers
of macro parameters for another token type and add their numbers
*/
static void 
pp_chain_subst(struct token* chain, struct pp_hash* hash)
{
	struct token* pt;
	struct pp_macro* pm;

	pt=chain->next;
	while(pt->type!=HEAD){
		if(pt->type==IDENTIFIER){
			if(pm=hash_find(pt->text,hash)){
				pt->type=PARAM;
				/* nasty hack: we are saving arg number
				in pt->gl because gl will will not be needed anyway */
				pt->gl=pm->arity;
			}	
		}	
		pt=pt->next;
	}
}

/*
saves new macro into defined_macros
*/
static void 
pp_save_define(char* c, int arity, struct token* chain, struct pp_hash* hash)
{
	struct pp_macro* new, * old;

	delete_token(chain->previous); /* delete last newline */	
	new=new_macro(c,arity,chain);
	old=hash_insert(new,hash);

	if (old){
 		if (arity!=old->arity || !token_chain_cmp(chain,old->chain))
			warn("Redefinition of a macro");
		free(old);
	}
#ifdef DEBUG
	printf("	pp_save_define: ");
	hash_show(hash);
	print_token_chain(chain);
	printf("\n");
#endif
}

/*
deletes macro from defined_macros
*/
static void 
pp_undef(char *c, struct pp_hash* hash)
{
	hash_delete(c,defined_macros);
#ifdef DEBUG
	hash_show(hash);
#endif
}

static struct token*
pp_parse_define(struct token* pt, struct token* head)
{
	char* name;
	int arity,control;
	struct pp_hash* hash;
	struct token* ret;

	ret=head;
	hash=hash_new();	

	pt=skip_whitespace(pt->next);

	if (pt->type!=IDENTIFIER)
		err("Bad define");
	name=pt->text;

	/* process part 2 (definition) */
	if (pt->next->type==LPAREN){
		arity=0;
		control=0;
		pt=pt->next;
		do {
			pt=skip_whitespace(pt->next);
			if (control<=0 && pt->type==IDENTIFIER){
				control=1;
				pp_save_arg(pt->text,arity,hash);
				arity++;
			} else if (control>=0 && (pt->type==RPAREN ||
			    !strcmp(pt->text,","))) {
				control=-1;
			} else {
				err("Syntax error");
			}
		} while(pt->type!=RPAREN);
		pt=skip_whitespace(pt->next);
	} else {
		arity=-1;
		pt=skip_whitespace(pt->next);
	}	
	/* process part 3 (replacement list) */
	if (pt->type!=HEAD) /* we do not want to get empty HEAD in pt */
		ret=split_chain(head,pt);
	else
		ret=new_chain(); /* empty definition: we will save empty chain */
	if (arity>0) //TODO: investigate why is this if necessary
		pp_chain_subst(ret,hash);

	hash_free(hash);
	pp_save_define(name,arity,ret,defined_macros);	

	return head;
}

static void subst(struct token* chain);

static struct token*
trim_directive_and_subst(struct token* chain)
{
	while (chain->next->type!=IDENTIFIER)
		delete_token(chain->next);
	delete_token(chain->next);
	subst(chain);
	return skip_whitespace(chain->next);
}

/*
parse pp directive - should be ran after substitution
*/
static struct token*
pp_parse_directive(struct token* head, struct token* hash)
{
	struct token* pt, * ret;

	ret=head;
	pt=hash;
	pt=skip_whitespace(pt->next);
	
	if (pt->type==HEAD)
		return head; /* empty directive */
	if (pt->type!=IDENTIFIER){
		err("Bad pp directive");
	}

	if (!strcmp(pt->text,"include")){
		pt=trim_directive_and_subst(ret);
		if ((pt->type!=FILENAME && pt->type!=STRING) || pt->next->type!=WHITESPACENL)
			err("Bad include");
		switch_file(extract_filename(pt->text),pt->type);
	} else if (!strcmp(pt->text,"define")){
		ret=pp_parse_define(pt,head);
	} else if (!strcmp(pt->text,"undef")) {
		pt=skip_whitespace(pt->next);
		if (pt->type==IDENTIFIER)
			pp_undef(pt->text,defined_macros);
		else
			err("Syntax error");
	}

	return ret;
}

/* prepares one chain for further parsing or substitution
   takes care for switching tokenizer mode on includes
*/
static struct token*
pp_chain(void)
{
	struct token* pt, *pc;
	
	pt=mk_token();
	if (!pt)
		return NULL;
	pc=new_chain();
	push_token(pt,pc);
	if (pt->type==HASH){ /* is it a directive? we read only a line then */
		if (!strcmp(skip_whitespace(pt->next)->text,"include"))/* switch tokenizer mode */
			tok_inc_enable=1;
		do {
			pt=mk_token();
			if (!pt)
				break;
			push_token(pt,pc);
		} while(pt->type!=WHITESPACENL);			
		tok_inc_enable=0;
	} else {
		do {
			pt=mk_token();
			push_token(pt,pc);
		} while(pt && (pt->type!=WHITESPACENL));			
	}
	return pc;
}

/* arity>=0 expected on input, pt should be LPAREN pointer */
static struct token**
parse_args(struct token* lparen, int arity)
{
	struct token** ret;
	int args_found, paren;
	struct token* pt, * lastcom;
#ifdef DEBUG
	printf("parse_args: Entering\n");
#endif

	pt=lastcom=lparen->next;
	args_found=0;
	paren=1;
	if (!(ret=malloc((arity ? arity : 1) * sizeof(struct token*))))
		err("Low memory");
	
	while (paren) {
		if (pt->type==HEAD){ /* we bumped the end of line and still do not have all parens */
			struct token* newline, * temp;

			strcpy(pt->previous->text," "); /* change last newline into space */
			if (!(newline=pp_chain())) /* ask for next line */
				err("Syntax error: Unclosed parens");
			temp=pt->previous; 
			insert_chain_after(pt->previous, newline);
			pt=temp->next;
		}
		if ((pt->type==OPERATOR && !(strcmp(pt->text,",")) || pt->type==RPAREN) && paren==1){
			ret[args_found++]=split_token_chain_ft(lastcom, pt);
#ifdef DEBUG
	printf("	parse_args: \n");
	print_token_chain(ret[args_found-1]);
	printf("\n");
#endif
			lastcom=pt->next;

		}
		if (args_found>arity)
			err("Too many arguments");

		if (pt->type==LPAREN)
			paren++;
		if (pt->type==RPAREN)
			paren--;
		pt=pt->next;
	}
	if (args_found!=arity)
		err("Not enough arguments");
#ifdef DEBUG
	printf("parse_args: leaving, arity=%d\n",arity);
#endif
	return ret;
	
}

/* and finally let us go to actual work ;) */
static void
subst(struct token* chain)
{
	struct token* pt, * pp, * ppt, * repl_list, * temp;
	struct pp_macro* pm;
	int arity, i;
	enum ttype temp_ttype; 
	struct token** args;

#ifdef DEBUG
	printf("subst: Entered subst\n");
#endif
	pt=chain->next;
	while (pt->type!=HEAD){
#ifdef DEBUG
	printf("       subst: We are working on %s\n",pt->text);
#endif

		if(pt->type==IDENTIFIER && (pm=hash_find(pt->text,defined_macros))){
#ifdef DEBUG
		printf("	subst: substituting macro %s\n",pt->text);
#endif
			arity=pm->arity;	
			pp=clone_token_chain(pm->chain);

			if (arity>=0){
				struct token *lpar = skip_whitespace(pt->next);
				if (lpar->type!=LPAREN)
					err("Arguments missing");
				args=parse_args(lpar,arity);
				for (i=0; i<arity; i++){
					subst(args[i]); //FIXME: one identifier should be substituted only once
#ifdef DEBUG
		printf("	subst: (arg chain %d)",i);
		print_token_chain(args[i]);
		printf("\n");
#endif
				}
				ppt=pp->next;
				while(ppt->type!=HEAD){
					if (ppt->type==PARAM){
#ifdef DEBUG
	printf("	subst: arg %d [%s]\n", ppt->gl,ppt->text);
#endif					
						repl_list=clone_token_chain(args[ppt->gl]);
						temp=ppt->next;
						insert_chain_after(ppt,repl_list);
						delete_token(ppt);
						ppt=temp;
					}	
					ppt=ppt->next;
				}
				for (i=0; i<arity; i++)
					free_chain(args[i]);
				free(args);

				do {
					temp_ttype=pt->next->type;
					delete_token(pt->next);
				} while (temp_ttype!=RPAREN);		
			}
			
			temp=pt->next;
			insert_chain_after(pt,pp);
			delete_token(pt);	
			pt=temp;
		}
		pt=pt->next;
	}	
#ifdef DEBUG
	printf("subst: Leaving\n");
#endif
}

/*
while we can get some tokens,
this one asks for a token chains and process them one by one
*/
static void
pp_go(void)
{
	struct token* chain, * p;	
	while(chain=pp_chain()){
		p=skip_whitespace(chain->next);
		if (p->type==HASH){
			/* this line is also substituted (according to 6.10.2)
			  directly in pp_parse_directive */
			chain=pp_parse_directive(chain,p);
		} else {
			subst(chain);	
			print_token_chain_plaintext(chain);
		}
		free_chain(chain);
	}
	
}

/* init global variables */
static void
hash_init(void)
{
	defined_macros=hash_new();
}

int
main(int argc, char **argv)
{
	if (argc<2){
		err("Filename not supplied.\nUsage: quork filename");
	}

	lex_init(argv[1]);
	hash_init();
	pp_go();
	
	return 0;
}
